import statistics
import seaborn as sns
import matplotlib.pyplot as plt
import pandas as pd

from .output_dir import OutputDir


class ValidationTrial(OutputDir):
    def __init__(self, runs, class_names=[]):
        self.runs = runs
        self.class_names = class_names

    def mean_set_size(self):
        means = []
        for run in self.runs:
            means.append(run.mean_set_size())
        return statistics.mean(means)

    def pct_empty_sets(self):
        empty = []
        for run in self.runs:
            empty.append(run.pct_empty_sets())
        return statistics.mean(empty)

    def pct_singleton_sets(self):
        singleton = []
        for run in self.runs:
            singleton.append(run.pct_singleton_sets())
        return statistics.mean(singleton)

    def pct_singleton_or_duo_sets(self):
        singleton_or_duo = []
        for run in self.runs:
            singleton_or_duo.append(run.pct_singleton_or_duo_sets())
        return statistics.mean(singleton_or_duo)

    def pct_trio_plus_sets(self):
        trio_plus = []
        for run in self.runs:
            trio_plus.append(run.pct_trio_plus_sets())
        return statistics.mean(trio_plus)

    def mean_false_negative_rate(self):
        fns = []
        for run in self.runs:
            fns.append(run.false_negative_rate())
        return statistics.mean(fns)

    def mean_model_false_negative_rate(self):
        fns = []
        for run in self.runs:
            fns.append(run.model_false_negative_rate())
        return statistics.mean(fns)

    def mean_true_positive_rate(self):
        tps = []
        for run in self.runs:
            tps.append(run.true_positive_rate())
        return statistics.mean(tps)

    def mean_model_true_positive_rate(self):
        tps = []
        for run in self.runs:
            tps.append(run.model_true_positive_rate())
        return statistics.mean(tps)

    def mean_softmax_threshold(self):
        return sum(run.softmax_threshold for run in self.runs) / len(self.runs)

    def mean_softmax_thresholds(self):
        return {key: statistics.mean([run.softmax_thresholds[key] for
                                      run in self.runs]) for key in self.runs[0].softmax_thresholds}

    def mean_set_sizes_by_class(self, class_names):
        set_size_dicts = []
        for run in self.runs:
            set_size_dicts.append(run.mean_set_sizes_by_class(class_names))

        means = {}
        for class_name in class_names:
            d_means = []
            for d in set_size_dicts:
                if class_name in d:
                    d_means.append(d[class_name])
            if len(d_means) > 0:
                means[class_name] = statistics.mean(d_means)
        return means

    def mean_fnrs_by_class(self, class_names):
        fnr_dicts = []
        for run in self.runs:
            fnr_dicts.append(run.mean_fnrs_by_class(class_names))

        means = {}
        for class_name in class_names:
            d_means = []
            for d in fnr_dicts:
                if class_name in d:
                    d_means.append(d[class_name])
            if len(d_means) > 0:
                means[class_name] = statistics.mean(d_means)
        return means

    def run_reports(self, base_output_dir, display_classes=None):
        self.create_output_dir(base_output_dir)
        self.visualize_empirical_fnr()
        self.visualize_class_performance(display_classes)
        print(f'Reports saved to {self.output_dir}')

    def visualize_empirical_fnr(self):
        plt.figure()

        # Generate a pastel palette
        color_palette = sns.color_palette("deep")

        ax = plt.gca()

        fnrs = []

        for run in self.runs:
            fnrs.append(run.false_negative_rate())

        # Get overall FNR
        mean_fnrs = round(self.mean_false_negative_rate(), 4)

        # Draw a histogram on the current subplot
        sns.histplot(fnrs, kde=False, ax=ax, color=color_palette[2])

        # Draw a vertical line at the mean
        ax.axvline(x=mean_fnrs, color='r', linestyle='--')

        # Set the labels of the x-axis and y-axis
        ax.set_xlabel('Mean run FNR')
        ax.set_ylabel('n runs')

        # Add a label to the inner right-hand corner of the current subplot
        ax.text(0.99, 0.95, f"Mean FNR across {len(self.runs)} runs: {mean_fnrs}", ha='right', va='top', transform=ax.transAxes)

        # Save the figure
        plt.savefig(f'{self.output_dir}/empirical_fnr.png')

    def visualize_class_performance(self, display_classes=None):
        def translate_class_name(class_name):
            if display_classes is None or class_name not in display_classes:
                return class_name
            else:
                return display_classes[class_name]

        def translate_dict(d):
            return {translate_class_name(key): value for key, value in d.items()}

        plt.figure()

        # FNRs by class
        fnrs_by_class = self.mean_fnrs_by_class(self.class_names)

        # Sort the dictionary by its values
        mean_fnrs = translate_dict(dict(sorted(fnrs_by_class.items(), key=lambda item: item[1])))

        # Visualize this dict as a bar chart
        sns.set_style('whitegrid')
        palette = sns.color_palette("deep")  # You can change "deep" to any other palette
        fig, ax = plt.subplots(figsize=(10, 6))
        ax.bar(range(len(mean_fnrs)), mean_fnrs.values(), color=palette[0])
        ax.set_xticks(range(len(mean_fnrs)))
        ax.set_xticklabels(mean_fnrs.keys(), rotation='vertical')
        ax.set_ylabel('Mean FNR')
        ax.set_xlabel('True class')
        plt.tight_layout()

        # Export as fig and text
        plt.savefig(f'{self.output_dir}/mean_fnrs_by_class.png')

        # Convert dictionary to dataframe
        df = pd.DataFrame(mean_fnrs, index=[0]).T

        # Transpose the dataframe
        df.to_csv(f'{self.output_dir}/mean_fnrs_by_class.csv',
                  index=True, header=False)

        # Reset plt
        plt.figure()

        # Set sizes by class
        mean_set_sizes = self.mean_set_sizes_by_class(self.class_names)

        # Sort the dictionary by its values
        mean_set_sizes = translate_dict(dict(sorted(mean_set_sizes.items(), key=lambda item: item[1])))

        # Convert dictionary to dataframe
        df = pd.DataFrame(mean_set_sizes, index=[0]).T

        # Transpose the dataframe
        df.to_csv(f'{self.output_dir}/mean_set_sizes_class.csv',
                  index=True, header=False)

        # Visualize this dict as a bar chart
        sns.set_style('whitegrid')
        fig, ax = plt.subplots(figsize=(10, 6))
        ax.bar(range(len(mean_set_sizes)), mean_set_sizes.values(), color=palette[1])
        ax.set_xticks(range(len(mean_set_sizes)))
        ax.set_xticklabels(mean_set_sizes.keys(), rotation='vertical')
        ax.set_ylabel('Mean set size')
        ax.set_xlabel('True class')
        plt.tight_layout()
        plt.savefig(f'{self.output_dir}/mean_set_sizes_by_class.png')



import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt

from .output_dir import OutputDir
from .prediction_dataset import PredictionDataset


class AlphaSelector(OutputDir):
    def __init__(self,
                 prediction_dataset: PredictionDataset,
                 cop_class,
                 base_output_dir,
                 min_alpha=0.05,
                 max_alpha=0.5,
                 increment_alpha=0.05,
                 n_runs_per_alpha=1000,
                 ):
        self.create_output_dir(base_output_dir)
        self.cop_class = cop_class
        self.prediction_dataset = prediction_dataset
        self.min_alpha = min_alpha
        self.max_alpha = max_alpha
        self.increment_alpha = increment_alpha
        self.n_runs_per_alpha = n_runs_per_alpha

        # Initialize lists
        self.alphas = []
        self.trials = {}
        self.mean_set_sizes = []
        self.pcts_empty_sets = []
        self.pcts_singleton_sets = []
        self.pcts_singleton_or_duo_sets = []
        self.pcts_trio_plus_sets = []
        self.mean_false_negative_rates = []
        self.mean_softmax_threshold = []
        self.lamhats = {}
        self.lambdas = {}

    def run(self):
        for a in np.arange(self.min_alpha,
                           self.max_alpha + self.increment_alpha,
                           self.increment_alpha):
            alpha = round(a, 2)
            cop = self.cop_class(self.prediction_dataset, alpha=alpha)
            trial = cop.do_validation_trial(n_runs=self.n_runs_per_alpha)

            # Store values for graphing
            self.alphas.append(alpha)
            self.trials[alpha] = trial
            self.mean_set_sizes.append(trial.mean_set_size())
            self.pcts_empty_sets.append(trial.pct_empty_sets())
            self.pcts_singleton_sets.append(trial.pct_singleton_sets())
            self.pcts_singleton_or_duo_sets.append(
                trial.pct_singleton_or_duo_sets())
            self.pcts_trio_plus_sets.append(trial.pct_trio_plus_sets())
            self.mean_false_negative_rates.append(
                trial.mean_false_negative_rate())
            self.mean_softmax_threshold.append(trial.mean_softmax_threshold())
            self.lamhats[alpha] = trial.mean_softmax_threshold()
            self.lambdas[alpha] = trial.mean_softmax_thresholds()

    def run_reports(self):
        self.visualize()
        self.visualize_lambdas()
        self.save_summary()
        print(f'Reports saved to {self.output_dir}')

    def visualize(self):
        # MEAN SET SIZES GRAPH
        plt.figure()

        data = pd.DataFrame({
            'Alpha': self.alphas,
            'Mean Set Size': self.mean_set_sizes
        })

        sns.lineplot(data=data, x='Alpha', y='Mean Set Size')
        plt.savefig(f'{self.output_dir}/alpha_to_mean_set_size.png')

        # PERCENT EMPTY/SINGLETON SETS GRAPH
        plt.figure()

        # Labels
        x_label = 'Alpha'
        y_label = 'Proportion of prediction sets with size n'
        legend_title = 'Set size'

        # Create a DataFrame for the pct_empty_sets and pct_singleton_sets
        data = pd.DataFrame({
            x_label: self.alphas,
            'n = 0': self.pcts_empty_sets,
            'n ∈ {1, 2}': self.pcts_singleton_or_duo_sets,
            'n ≥ 3': self.pcts_trio_plus_sets
        })

        # Melt the DataFrame to have the set types as a separate column
        data_melted = data.melt(id_vars=x_label,
                                var_name=legend_title,
                                value_name=y_label)

        # Create the bar chart
        sns.barplot(data=data_melted,
                    x=x_label,
                    y=y_label,
                    hue=legend_title,
                    alpha=0.5)

        # Get the current x-tick labels
        labels = [item.get_text() for item in plt.gca().get_xticklabels()]

        # Draw a horizontal line across the top of the highest orange bar
        max_singleton_or_duo_sets = data['n ∈ {1, 2}'].max()
        plt.axhline(y=max_singleton_or_duo_sets,
                    color='#cccccc',
                    linestyle='--')

        # Get the index of the label with the highest value
        idx = data['n ∈ {1, 2}'].idxmax()

        # Make this label bold
        labels[idx] = f'$\\bf{{{labels[idx]}}}$'
        # Set the new x-tick labels
        plt.gca().set_xticklabels(labels)

        # Add a legend
        plt.legend(loc='upper right', title=legend_title)

        # Save the plot to a file
        plt.savefig(f'{self.output_dir}/alpha_to_set_sizes.png')

    def visualize_lambdas(self):
        plt.figure()

        # Only use reasonable alphas
        alphas = [0.05, 0.1, 0.15, 0.2, 0.3, 0.4]

        SYM_LAMHAT = r'$\hatλ$'
        n_colors = len(alphas)
        # Select n_colors colors from the 'tab10' colormap
        colors = sns.color_palette('tab10', n_colors)

        i = 0
        for a in alphas:
            # Plot lamdas as lineplot with each alpha in different color
            sns.lineplot(x=list(self.lambdas[a].keys()),
                         y=list(self.lambdas[a].values()),
                         color=colors[i],
                         label=f'α={a}')

            # Draw a circle at the lambda[a] value on x and 0 on y
            plt.scatter(self.lamhats[a], 0, s=100, facecolors='none', edgecolors='black')

            padding = 0.05  # adjust this value to get the desired amount of space
            plt.text(self.lamhats[a], 0 + padding,
                     f'{self.lamhats[a]:.2f}',
                     ha='center', va='bottom',
                     fontsize=8, color='black',
                     weight='bold')
            i += 1

        plt.legend(title='Error rate')

        # Draw a dashed gray line at 0
        plt.axhline(0, color='gray', linestyle='--')

        plt.xlabel('λ')
        plt.ylabel('Constraint function f(λ)')
        plt.title(f'FNR-controlling threshold {SYM_LAMHAT}')

        # Save the plot to a file
        plt.savefig(f'{self.output_dir}/lambdas.png')

    def save_summary(self):
        # Export the statistics to CSV
        # Create a dictionary where the keys are the column names and the values are the lists
        data = {
            'alpha': self.alphas,
            'Mean set size': self.mean_set_sizes,
            '% sets n=0': self.pcts_empty_sets,
            '% sets n={1}': self.pcts_singleton_sets,
            '% sets n={1|2}': self.pcts_singleton_or_duo_sets,
            '% sets n>=3': self.pcts_trio_plus_sets,
            'Mean FNR': self.mean_false_negative_rates,
            'Mean softmax threshold': self.mean_softmax_threshold
        }

        # Create a DataFrame from the dictionary
        df = pd.DataFrame(data)

        # Export the DataFrame to a CSV file
        df.to_csv(f'{self.output_dir}/alpha_selection.csv', index=False)



import numpy as np
from .base_cop import BaseCoP
from .prediction_dataset import PredictionDataset


class SimpleCoP(BaseCoP):
    def __init__(self, prediction_dataset: PredictionDataset, alpha=0.1):
        super().__init__(prediction_dataset, alpha)

    def calibrate(self):
        super().calibrate()

        # Conformal prediction time
        # 1: get conformal scores.
        cal_scores = 1 - self.cal_smx[np.arange(self.n_cal), self.cal_labels]

        # 2: get adjusted quantile
        q_level = np.ceil((self.n_cal + 1) * (1 - self.alpha)) / self.n_cal
        self.qhat = np.quantile(cal_scores, q_level, interpolation='higher')
        self.softmax_threshold = 1 - self.qhat



import numpy as np
import pandas as pd
from .output_dir import OutputDir


class ValidationRun(OutputDir):
    EPSILON = 1e-10

    def __init__(self,
                 smx_idx,
                 sample_names,
                 prediction_sets,
                 labels_idx,
                 model_predictions,
                 softmax_threshold,
                 class_names,
                 softmax_thresholds=None):
        # The indices in the softmax array that correspond to the
        # samples in this run
        self.smx_idx = smx_idx

        # The names of the samples in this run
        self.sample_names = sample_names

        # The predictions for this run
        self.prediction_sets = prediction_sets

        # The labels for all samples as indices
        self.labels_idx = labels_idx

        # The qhat/lamhat value for this run, if provided
        self.softmax_threshold = softmax_threshold

        # The model predictions for this run
        self.model_predictions = model_predictions

        # The class names
        self.class_names = class_names

        # All attempted softmax thresholds
        self.softmax_thresholds = softmax_thresholds

    def mean_set_size(self):
        return sum(sum(prediction_set) for
                   prediction_set in self.prediction_sets) / \
                    len(self.prediction_sets)

    def pct_empty_sets(self):
        return sum(sum(prediction_set) == 0 for
                   prediction_set in self.prediction_sets) / \
                    len(self.prediction_sets)

    def pct_singleton_sets(self):
        return sum(sum(prediction_set) == 1 for
                   prediction_set in self.prediction_sets) / \
                    len(self.prediction_sets)

    def pct_singleton_or_duo_sets(self):
        return sum(sum(prediction_set) == 1 or sum(prediction_set) == 2 for
                   prediction_set in self.prediction_sets) / \
                    len(self.prediction_sets)

    def pct_trio_plus_sets(self):
        return sum(sum(prediction_set) >= 3 for
                   prediction_set in self.prediction_sets) / \
                    len(self.prediction_sets)

    def false_negative_rate(self):
        return 1 - ((self.prediction_sets * self.labels_idx).sum(axis=1) /
                    (self.labels_idx.sum(axis=1) + self.EPSILON)).mean()

    def model_false_negative_rate(self):
        return 1 - ((self.model_predictions * self.labels_idx).sum(axis=1) /
                    (self.labels_idx.sum(axis=1) + self.EPSILON)).mean()

    def true_positive_rate(self):
        return ((self.prediction_sets * self.labels_idx).sum(axis=1) /
                (self.labels_idx.sum(axis=1) + self.EPSILON)).mean()

    def model_true_positive_rate(self):
        return ((self.model_predictions * self.labels_idx).sum(axis=1) /
                (self.labels_idx.sum(axis=1) + self.EPSILON)).mean()

    def bin_set_sizes_by_class(self, class_names):
        set_sizes = {}
        for i in range(len(self.prediction_sets)):
            labels = self.labels_idx[i]
            # Get corresponding values from class_names
            pset_class_names = [class_names[i] for i, label in enumerate(labels) if label == 1]
            pset = self.prediction_sets[i]
            for class_name in pset_class_names:
                sizes = set_sizes.get(class_name, [])
                sizes.append(sum(pset))
                set_sizes[class_name] = sizes
        return set_sizes

    def mean_set_sizes_by_class(self, class_names):
        set_sizes = self.bin_set_sizes_by_class(class_names)
        averages = {}
        for key in set_sizes:
            sizes = set_sizes[key]
            averages[key] = sum(sizes) / len(sizes)
        return averages

    def mean_fnrs_by_class(self, class_names):
        fnrs = {}
        for i in range(len(self.prediction_sets)):
            labels = self.labels_idx[i]
            # Get corresponding values from class_names
            pset_class_names = [class_names[i] for i, label in enumerate(labels) if label == 1]
            pset = np.array([int(value) for value in self.prediction_sets[i]])
            if (pset[labels == 1] == 1).size > 0:
                fnr = 1 - np.mean((pset[labels == 1] == 1))
            else:
                fnr = np.nan  # or any other value you want to assign when the array is empty
            for class_name in pset_class_names:
                class_fnrs = fnrs.get(class_name, [])
                class_fnrs.append(fnr)
                fnrs[class_name] = class_fnrs
        averages = {}
        for key in fnrs:
            sizes = fnrs[key]
            averages[key] = sum(sizes) / len(sizes)
        return averages

    def run_reports(self, base_output_dir):
        np.seterr(all='raise')
        self.create_output_dir(base_output_dir)

        df = pd.DataFrame({
            'false_negative_rate': self.false_negative_rate(),
            'model_false_negative_rate': self.model_false_negative_rate(),
            'mean_set_size': self.mean_set_size(),
            'pct_empty_sets': self.pct_empty_sets(),
            'pct_singleton_sets': self.pct_singleton_sets(),
            'pct_singleton_or_duo_sets': self.pct_singleton_or_duo_sets(),
            'pct_trio_plus_sets': self.pct_trio_plus_sets()
        }, index=[0])

        df.T.to_csv(f'{self.output_dir}/summary.csv', header=False)

        df = pd.DataFrame(self.mean_set_sizes_by_class(self.class_names), index=[0])
        df.T.to_csv(f'{self.output_dir}/mean_set_sizes_by_class.csv', header=False)

        df = pd.DataFrame(self.mean_fnrs_by_class(self.class_names), index=[0])
        df.T.to_csv(f'{self.output_dir}/mean_fnrs_by_class.csv', header=False)

        print(f'Reports saved to {self.output_dir}')



import pandas as pd
import numpy as np
from matplotlib import pyplot as plt
import seaborn as sns

from .output_dir import OutputDir


class PredictionDataset(OutputDir):
    DATASET_NAME_COL = 'dataset'
    ID_COL = 'id'
    KNOWN_CLASS_COL = 'known_class'
    PREDICTED_CLASS_COL = 'predicted_class'
    MELTED_KNOWN_CLASS_COL = 'melted_known_class'

    def __init__(self,
                 df=None,
                 predictions_csv=None,
                 dataset_col_name=None,
                 dataset_name=None,
                 display_classes=None):
        self.output_dir = None

        if df is None and predictions_csv is None:
            raise ValueError('Either df or predictions_csv must be provided')

        if df is None:
            df = pd.read_csv(predictions_csv)

        self.df = df

        # Fill na's with 0s for prediction columns only
        self.df[self.class_names()] = self.df[self.class_names()].fillna(0)
        if dataset_col_name:
            self._create_dataset_column_from_col(dataset_col_name)
        elif dataset_name:
            self._create_dataset_column_from_name(dataset_name)
        self._parse_predictions()

        self.display_classes = display_classes

    def _create_dataset_column(self):
        # make dataset the first column
        self.df = self.df[[self.DATASET_NAME_COL] +
                          [col for col in self.df.columns
                           if col != self.DATASET_NAME_COL]]

        # De-fragment
        self.df = self.df.copy()

    def _create_dataset_column_from_col(self, col_name):
        self.df[self.DATASET_NAME_COL] = self.df[col_name]
        self._create_dataset_column()

    def _create_dataset_column_from_name(self, name):
        self.df[self.DATASET_NAME_COL] = name
        self._create_dataset_column()

    def _parse_predictions(self):
        self.smx = self.df[self.class_names()].values

        # Replace 1.0 with 0.99999
        self.smx[self.smx == 1.0] = 1 - (1e-10)

        self.predictions_str = self.df[
            PredictionDataset.PREDICTED_CLASS_COL].values

        # Replace nans with empty strings
        self.predictions_str = np.where(pd.isnull(self.predictions_str),
                                        '', self.predictions_str)

        predicted_classes = [classes.split(',') for classes
                             in self.predictions_str]

        self.predictions_bool = np.array(
            [[class_name in labels for class_name in self.class_names()]
             for labels in predicted_classes])

        self.labels_str = None
        self.labels_idx = None
        if self.KNOWN_CLASS_COL not in self.df.columns:
            return

        self.labels_str = self.df[
            PredictionDataset.KNOWN_CLASS_COL].values

        class_names = self.class_names()
        known_classes = [classes.split(',') for classes
                         in self.df[self.KNOWN_CLASS_COL].values]

        indices_list = [[class_names.index(class_name)
                         for class_name in class_list if class_name in
                         class_names] for class_list in known_classes]

        # convert the list of indices to a binary list
        self.labels_idx = np.array([[1 if i in indices else 0
                                    for i in range(len(self.class_names()))]
                                    for indices in indices_list])

    def append_dataset(self, other):
        self.df = pd.concat([self.df, other.df])

    def export(self, path):
        self.df.to_csv(path, index=False)

    # Create a new df creating a new record for every known class
    def melt(self):
        # Take KNOWN_CLASS_COL and split the values by comma into a new df with a column for each class
        known_classes_df = self.df[self.KNOWN_CLASS_COL].str.split(',', expand=True)

        # Label each column as known_class_1, known_class_2, etc.
        known_classes_df.columns = [f'{self.KNOWN_CLASS_COL}_{i+1}' for
                                    i in range(known_classes_df.shape[1])]

        df = pd.concat([self.df, known_classes_df], axis=1)

        # Use melt to reshape the DataFrame
        new_df = df.melt(id_vars=self.ID_COL,
                         value_vars=known_classes_df.columns,
                         value_name=self.MELTED_KNOWN_CLASS_COL)

        # Drop the 'variable' column and any rows with null 'class' values
        new_df = new_df.drop(columns='variable').dropna(subset=[self.MELTED_KNOWN_CLASS_COL])

        # Re-join it with the original DataFrame
        new_df = new_df.merge(self.df, left_on=self.ID_COL, right_on=self.ID_COL)

        return new_df

    def class_counts(self, translate=False):
        counting_df = self.melt()[self.MELTED_KNOWN_CLASS_COL]
        counts = counting_df.value_counts()
        if translate and self.display_classes:
            return counts.rename(index=self.display_classes)
        # Remove index name
        counts.index.name = None
        return counts

    def translate_class_name(self, class_name):
        if self.display_classes and class_name in self.display_classes:
            return self.display_classes[class_name]
        return class_name

    def class_names(self, translate=False):
        cols_to_exclude = [self.DATASET_NAME_COL, self.ID_COL,
                           self.KNOWN_CLASS_COL, self.PREDICTED_CLASS_COL,
                           self.MELTED_KNOWN_CLASS_COL]

        cols = [col for col in self.df.columns
                if col not in cols_to_exclude]

        if translate and self.display_classes:
            return [self.translate_class_name(col) for col in cols]

        # Return everything that is not in the exclusion list
        return cols

    def run_reports(self, base_output_dir):
        self.create_output_dir(base_output_dir)
        self.visualize_class_counts()
        self.visualize_prediction_heatmap()
        print(f'Reports saved to {self.output_dir}')

    def visualize_class_counts(self):
        plt.figure()

        # create a bar chart
        ccs = self.class_counts()

        # Translate if necessary
        if self.display_classes:
            ccs = ccs.rename(index=self.display_classes)

        # Print count above each bar
        for i, v in enumerate(ccs):
            plt.text(i, v, str(v), ha='center', va='bottom')

        ccs.plot.bar()

        # Dump class counts to CSV
        ccs.to_csv(f'{self.output_dir}/class_counts.csv')

        # show the plot
        plt.savefig(f'{self.output_dir}/class_counts.png', bbox_inches='tight')

    def visualize_prediction_heatmap(self):
        plt.figure(figsize=(10, 8))

        group_by_col = self.MELTED_KNOWN_CLASS_COL
        df = self.melt()

        grouped_df = df.groupby(group_by_col)
        pred_col_names = self.class_names()

        mean_smx = []

        for name, group in grouped_df:
            name = self.translate_class_name(name)
            mean_smx_row = [name]

            for col in pred_col_names:
                mean_smx_row.append(group[col].mean())

            mean_smx.append(mean_smx_row)

        col_names = ['true_class_name'] + self.class_names(translate=True)

        mean_smx_df = pd.DataFrame(mean_smx, columns=col_names)
        mean_smx_df.set_index('true_class_name', inplace=True)

        # Sort the rows and columns
        mean_smx_df.sort_index(axis=0, inplace=True)  # Sort rows
        mean_smx_df.sort_index(axis=1, inplace=True)  # Sort columns

        # Remove any columns where all the rows are 0
        mean_smx_df = mean_smx_df.loc[:, (mean_smx_df != 0).any(axis=0)]

        hm = sns.heatmap(mean_smx_df,
                 cmap="coolwarm",
                 annot=True,
                 fmt='.2f')

        labelpad = 20
        plt.setp(hm.get_yticklabels(), rotation=0)

        hm.set_xlabel('MEAN PROBABILITY SCORE',
                                 weight='bold', labelpad=labelpad)
        hm.set_ylabel('TRUE CLASS',
                                 weight='bold', labelpad=labelpad)

        # Save the plot to a file
        plt.savefig(f'{self.output_dir}/prediction_heatmap.png', bbox_inches='tight')

    def prediction_sets_df(self, prediction_sets, export_to_dir=None):
        # Make a copy of the DataFrame
        df = self.df.copy()

        # Add the prediction sets to the DataFrame
        df['prediction_sets'] = prediction_sets

        has_known_class = self.KNOWN_CLASS_COL in df.columns

        if has_known_class:
            # Get the known class names of the prediction set members
            def process_kc_row(row):
                classes = []
                for col in str(row[self.KNOWN_CLASS_COL]).split(','):
                    if col in row:
                        classes.append(str(row[col]))
                return ','.join(classes)

            df['known_class_softmax_scores'] = df.apply(
                lambda row: process_kc_row(row), axis=1)

        # Get the softmax scores of the prediction set members
        def process_row(row):
            scores = []
            if row['prediction_sets'] is None:
                return ''
            for col in str(row['prediction_sets']).split(','):
                if col in row:
                    scores.append(str(row[col]))
            return ','.join(scores)

        df['prediction_set_softmax_scores'] = df.apply(
            lambda row: process_row(row), axis=1)

        cols_to_keep = [self.DATASET_NAME_COL, self.ID_COL,
                        self.PREDICTED_CLASS_COL,
                        'prediction_sets']

        if has_known_class:
            cols_to_keep.append(self.KNOWN_CLASS_COL)

        cols_to_keep.append('prediction_set_softmax_scores')

        if has_known_class:
            cols_to_keep.append('known_class_softmax_scores')

        df = df[cols_to_keep]

        # Export the DataFrame to a CSV file
        if export_to_dir:
            if not self.output_dir:
                self.create_output_dir(export_to_dir)
            df.to_csv(f'{self.output_dir}/prediction_sets.csv', index=False)
            print(f'Prediction sets saved to {self.output_dir}/prediction_sets.csv') # noqa

        return df




from src.conformist.prediction_dataset import PredictionDataset
from src.conformist.fnr_cop import FNRCoP
import os
import shutil

# Get current directory
CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
OUTPUT_DIR = f'{CURRENT_DIR}/output'
INPUT_CSV = f'{CURRENT_DIR}/cal.csv'


def _cleanup():
    print(f'Cleaning up {OUTPUT_DIR}')
    # Remove all files and dirs in the output directory
    for filename in os.listdir(OUTPUT_DIR):
        file_path = os.path.join(OUTPUT_DIR, filename)
        if os.path.isfile(file_path) or os.path.islink(file_path):
            os.remove(file_path)
        elif os.path.isdir(file_path):
            shutil.rmtree(file_path)


def test_prediction_dataset():
    # Create dummy output directory if it doesn't exist
    if not os.path.exists(OUTPUT_DIR):
        os.makedirs(OUTPUT_DIR)

    try:
        calpd = PredictionDataset(predictions_csv=INPUT_CSV,
                                  dataset_name='my_cal_data')

        calpd.run_reports(OUTPUT_DIR)
    finally:
        _cleanup()


def test_fnr_cop():
    # Create dummy output directory if it doesn't exist
    if not os.path.exists(OUTPUT_DIR):
        os.makedirs(OUTPUT_DIR)

    try:
        calpd = PredictionDataset(predictions_csv=INPUT_CSV,
                                  dataset_name='my_cal_data')
        # Validation trial and reports
        mcp = FNRCoP(calpd, alpha=0.2)
        trial = mcp.do_validation_trial(n_runs=10)
        trial.run_reports(OUTPUT_DIR)
    finally:
        print("hi")
        #_cleanup()



import numpy as np
from scipy.optimize import brentq

from .base_cop import BaseCoP
from .prediction_dataset import PredictionDataset


class FNRCoP(BaseCoP):
    def __init__(self, prediction_dataset: PredictionDataset, alpha=0.1):
        super().__init__(prediction_dataset, alpha)

    def false_negative_rate(self, prediction_set, gt_labels):
        return 1 - ((prediction_set * gt_labels).sum(axis=1) /
                    gt_labels.sum(axis=1)).mean()

    def calibrate(self):
        super().calibrate()

        def lamhat_threshold(lam):
            return self.false_negative_rate(
                    self.cal_smx >= lam, self.cal_labels) - \
                        ((self.n_cal + 1) / self.n_cal * self.alpha - 1
                         / (self.n_cal + 1))

        self.lamhat = brentq(lamhat_threshold, 0, 1)
        self.softmax_threshold = self.lamhat

        # For graphing
        # lambdas to test: from 0.0 to 1.0 in 0.01 increments
        self.lambdas = np.linspace(0, 1, 101)
        self.lambdas = {lam: lamhat_threshold(lam) for lam in self.lambdas}
        self.softmax_thresholds = self.lambdas



import os
import shutil
from datetime import datetime


class OutputDir():
    def create_output_dir(self, base_output_dir='output', class_suffix=None):
        # Output directory should be output dir plus class name plus timestamp
        timestamp = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')

        name = self.__class__.__name__
        if class_suffix:
            name = f'{name}_{class_suffix}'

        self.output_dir = os.path.join(base_output_dir,
                                       name,
                                       timestamp)
        os.makedirs(self.output_dir)

    def export(self, filename, dst):
        src = f'{self.output_dir}/{filename}'
        os.makedirs(dst, exist_ok=True)
        shutil.copy(src, dst)
        print(f"Exported to {dst}/{filename}")



import matplotlib.pyplot as plt
import seaborn as sns
from . import OutputDir, PredictionDataset


class ModelVsCopFNR(OutputDir):
    def __init__(self,
                 prediction_dataset: PredictionDataset,
                 cop_class,
                 base_output_dir,
                 n_runs_per_alpha=10000,
                 alphas=[0.1, 0.2, 0.3],
                 colors=['blue', 'purple', 'green']
                 ):
        self.create_output_dir(base_output_dir)
        self.cop_class = cop_class
        self.prediction_dataset = prediction_dataset
        self.n_runs_per_alpha = n_runs_per_alpha
        self.alphas = alphas
        self.colors = colors

    def run(self):
        def do_round(x):
            return round(x, 3)

        self.trial_results = {}

        for alpha in self.alphas:
            cop = self.cop_class(self.prediction_dataset, alpha=alpha)
            trial = cop.do_validation_trial(n_runs=self.n_runs_per_alpha)

            mean_cp_fnr = do_round(trial.mean_false_negative_rate())
            mean_model_fnr = do_round(trial.mean_model_false_negative_rate())

            mean_cp_tpr = do_round(trial.mean_true_positive_rate())
            mean_model_tpr = do_round(trial.mean_model_true_positive_rate())

            cp_fnrs = []
            cp_tprs = []

            model_fnrs = []
            model_tprs = []

            for run in trial.runs:
                cp_fnrs.append(run.false_negative_rate())
                cp_tprs.append(run.true_positive_rate())
                model_fnrs.append(run.model_false_negative_rate())
                model_tprs.append(run.model_true_positive_rate())

            self.trial_results[alpha] = {'mean_cp_fnr': mean_cp_fnr,
                                         'mean_model_fnr': mean_model_fnr,
                                         'mean_cp_tpr': mean_cp_tpr,
                                         'mean_model_tpr': mean_model_tpr,
                                         'cp_fnrs': cp_fnrs,
                                         'cp_tprs': cp_tprs,
                                         'model_fnrs': model_fnrs,
                                         'model_tprs': model_tprs}

    def run_reports(self):
        plt.figure()

        model_rates = []
        model_mean = 0
        i = 0

        for alpha in self.alphas:
            color = self.colors[i]
            tr = self.trial_results[alpha]
            cp_rates = tr['cp_fnrs']
            model_rates = tr['model_fnrs']
            cp_mean = tr['mean_cp_fnr']
            model_mean = tr['mean_model_fnr']

            # Create a KDE plot of cp_fnrs and model_fnrs
            sns.kdeplot(cp_rates, label=f'CP, α={alpha}', color=color)

            # Draw lines for mean_cp_fnr and mean_model_fnr
            plt.axvline(cp_mean, color=color, linestyle='dashed', linewidth=2)

            # Add text labels for mean_cp_fnr and mean_model_fnr
            plt.text(cp_mean - 0.01, plt.ylim()[1]*0.95, f'x\u0304={cp_mean}', color=color, ha='right', weight='bold')
            i += 1

        sns.kdeplot(model_rates, label='Model', color='orange')
        plt.axvline(model_mean, color='orange', linestyle='dashed', linewidth=2)
        plt.text(model_mean + 0.01, plt.ylim()[1]*0.85, f'x\u0304={model_mean}', color='orange', ha='left', weight='bold')

        # Add labels and legend
        plt.xlabel('False Negative Rates')
        plt.ylabel('Frequency')
        plt.legend(loc='upper right')

        # Save the plot to a file
        plt.savefig(f'{self.output_dir}/model_vs_CoP_FNR.png', bbox_inches='tight')
        print(f'Reports saved to {self.output_dir}')



import math
import numpy as np
import matplotlib.pyplot as plt
from upsetplot import plot
import pandas as pd

from .validation_run import ValidationRun
from .validation_trial import ValidationTrial
from .prediction_dataset import PredictionDataset
from .output_dir import OutputDir


DEFAULT_VALIDATION_PROPORTION = 0.25


class BaseCoP(OutputDir):
    def __init__(self, prediction_dataset: PredictionDataset, alpha=0.1):
        self.prediction_dataset = prediction_dataset
        self.alpha = alpha
        self.class_names = prediction_dataset.class_names()
        self._df = prediction_dataset.df
        self.n_total = self._df.shape[0]

        # Initialize
        self.n_cal, self.cal_smx, self.cal_labels, self.cal_labels_idx = \
            None, None, None, None

        # Depending on the type of predictor, this will be e.g. qhat or lamhat
        self.softmax_threshold = None

        # This is all the thresholds attempted by the CoP
        self.softmax_thresholds = None

        # Parse predictions
        self.smx = self.prediction_dataset.smx
        self.model_predictions = self.prediction_dataset.predictions_bool
        self.model_predictions_txt = self.prediction_dataset.predictions_str
        self.labels_txt = self.prediction_dataset.labels_str

    def calibrate(self):
        if self.n_cal is None or self.cal_smx is None or self.cal_labels is None:
            self._set_calibration_index()

    def get_prediction_sets(self):
        return self.val_smx >= self.softmax_threshold

    def _set_calibration_index(self, shuffle=False):
        if self.n_cal is None:
            self.n_cal = self.n_total

        idx = np.array([1] *
                       self.n_cal + [0] *
                       (self.smx.shape[0] - self.n_cal)) > 0
        if shuffle:
            np.random.shuffle(idx)

        # Calibration data
        self.cal_smx = self.smx[idx, :]
        self.cal_labels = self.prediction_dataset.labels_idx[idx]

        return idx

    def split_data(self,
                   validation_proportion,
                   shuffle=False):

        # Determine number of calibration and validation data points
        if validation_proportion == 0:
            self.n_cal = self.n_total
            self.n_val = 0
        else:
            calibration_proportion = 1 - validation_proportion
            self.n_cal = math.ceil(self.n_total * calibration_proportion)
            self.n_val = self.n_total - self.n_cal

        idx = self._set_calibration_index(shuffle)

        # Validation data
        if self.n_val > 0:
            self.val_smx = self.smx[~idx, :]
            self.val_labels = self.prediction_dataset.labels_idx[~idx]
            self.val_sample_names = np.array(self._df[
                PredictionDataset.ID_COL].values)[~idx]
            self.val_labels_txt = self.labels_txt[~idx]
            self.val_model_predictions = self.model_predictions[~idx]
            self.val_idx = ~idx

    def shuffle_data(self, validation_proportion):
        self.split_data(validation_proportion, shuffle=True)

    def do_validation_run(self, validation_proportion):
        self.shuffle_data(validation_proportion)
        self.calibrate()
        return ValidationRun(
                self.val_idx,
                self.val_sample_names,
                self.get_prediction_sets(),
                self.val_labels,
                self.val_model_predictions,
                self.softmax_threshold,
                self.class_names,
                self.softmax_thresholds)

    def do_validation_trial(self,
                            n_runs=1,
                            val_proportion=DEFAULT_VALIDATION_PROPORTION):
        self.runs = []
        for i in range(n_runs):
            result = self.do_validation_run(val_proportion)
            self.runs.append(result)
        return ValidationTrial(self.runs, self.class_names)

    def predict(self,
                pds,
                export_to_dir=None,
                validate=False,
                display_classes=None,
                upset_plot_color="black"):
        self.smx = pds.smx
        self.val_smx = pds.smx
        self.val_labels = pds.labels_idx
        self.val_sample_names = pds.df[PredictionDataset.ID_COL].values
        self.val_labels_txt = pds.labels_str
        self.val_model_predictions = pds.predictions_bool
        self.val_idx = np.array([1] * pds.smx.shape[0])

        prediction_sets = self.get_prediction_sets()
        prediction_sets_text = self.prediction_sets_to_text(prediction_sets)
        formatted_predictions = pds.prediction_sets_df(prediction_sets_text,
                                                       export_to_dir)

        if export_to_dir:
            self.upset_plot(
                prediction_sets,
                export_to_dir,
                display_classes,
                upset_plot_color)

        if not validate:
            return formatted_predictions

        vr = ValidationRun(
                self.val_idx,
                self.val_sample_names,
                prediction_sets,
                self.val_labels,
                self.val_model_predictions,
                self.softmax_threshold,
                self.class_names)

        if export_to_dir:
            vr.run_reports(export_to_dir)

        return formatted_predictions, vr


    def prediction_set_to_text(self, prediction_set, display_classes=None):
        class_names = self.class_names

        # Replace any class_names in display_classes with the display name
        if display_classes:
            class_names = [display_classes.get(cn, cn) for cn in class_names]

        ps = [class_names[i] for i in range(len(prediction_set)) if
              prediction_set[i]]

        return ','.join(ps)

    def prediction_sets_to_text(self, prediction_sets, display_classes=None):
        return [self.prediction_set_to_text(prediction_set, display_classes)
                for prediction_set in prediction_sets]

    def upset_plot(self, predictions_sets, export_to_dir, display_classes=None, color="black"):
        plt.figure()

        class_names = self.class_names
        # Replace any class_names in display_classes with the display name
        if display_classes:
            class_names = [display_classes.get(cn, cn) for cn in class_names]

        # Make prediction_sets into a df with class_names as columns
        upset_data = pd.DataFrame(predictions_sets, columns=class_names)

        upset_data.fillna(0, inplace=True)

        # Remove columns with no data
        upset_data = upset_data.loc[:, upset_data.sum(axis=0) > 0]

        # Set a multi-index
        upset_data.set_index(upset_data.columns.tolist(), inplace=True)

        plot(upset_data, sort_by="cardinality", facecolor=color, show_counts="%d", show_percentages="{:.0%}")

        path = f'{export_to_dir}/upset_plot.png'
        print("Saving upset plot to", path)
        plt.savefig(path)



from .alpha_selector import AlphaSelector
from .fnr_cop import FNRCoP
from .output_dir import OutputDir
from .prediction_dataset import PredictionDataset
from .prediction_data_preprocessor import PredictionDataPreprocessor
from .validation_trial import ValidationTrial
from .validation_run import ValidationRun
from .model_vs_cop_fnr import ModelVsCopFNR



import pandas as pd
from . import OutputDir, PredictionDataset


class PredictionDataPreprocessor(OutputDir):
    def __init__(self,
                 model_name,
                 predictions_csv,
                 dataset_name,
                 id_col,
                 predicted_class_col,
                 known_classes_df=None,
                 csv_delimiter=','):

        self.model_name = model_name
        self.df_raw = pd.read_csv(predictions_csv, delimiter=csv_delimiter)
        self.predicted_class_col = predicted_class_col
        self.set_name(dataset_name)

        # Start building the processed df
        self.df = pd.DataFrame()
        self.df[PredictionDataset.ID_COL] = self.df_raw[id_col]
        self.df[PredictionDataset.DATASET_NAME_COL] = dataset_name

        self.df[PredictionDataset.PREDICTED_CLASS_COL] = \
            self.df_raw[predicted_class_col]

        if known_classes_df is not None:
            subframe = known_classes_df[[PredictionDataset.KNOWN_CLASS_COL,
                                         PredictionDataset.ID_COL]]
            self.df = pd.merge(self.df,
                               subframe,
                               on=PredictionDataset.ID_COL)

    def set_name(self, dataset_name):
        self.dataset_name = dataset_name
        self.output_file = f'{self.model_name}_predictions_{dataset_name}.csv'

    def save(self, base_output_dir):
        self.create_output_dir(base_output_dir, self.dataset_name)
        path = f'{self.output_dir}/{self.output_file}'
        self.df.to_csv(path, index=False)
        print(f"Saved to {path}")

    def export(self, dst):
        super().export(self.output_file, dst)

    def append_dataset(self, other, merged_name=None):
        if merged_name:
            self.set_name(merged_name)
        self.df = pd.concat([self.df, other.df])
